# -*- coding: utf-8 -*-

"""
@date: 2022/4/4 上午11:18
@file: key_word.py
@author: zj
@description: Common keywords
"""

# Output identifier
KEY_OUTPUT = 'output'
# Loss identifier
KEY_LOSS = 'loss'

# Separator for csv file
KEY_SEP = ',,'
# Data file name is used for MPDataset
KEY_DATASET = 'data.csv'
# Class file name is used for MPDataset
KEY_CLASSES = 'cls.csv'

# empty identifier
KEY_EMPTY = ""
